from typing import List

from tortoise import Tortoise

from propan.db.workers.model.connection_params import Connection


async def init_tortoise(conn: Connection, ways_to_models: List[str]):
    await Tortoise.init(
        db_url=f'postgres://{conn.user}:{conn.password}@{conn.host}/{conn.name}',
        modules={'models': ways_to_models}
    )



from functools import wraps
from typing import Tuple

from .model.usecase import LoggerUsecase


class LoggerSimpleComposition(LoggerUsecase):
    loggers: Tuple[LoggerUsecase]
    not_catch: Tuple[Exception]

    def __init__(
        self,
        *loggers: Tuple[LoggerUsecase, int],
        not_catch: Tuple[Exception] = tuple()
    ):
        self.loggers = loggers
        self.not_catch = not_catch
        for logger in self.loggers:
            logger.not_catch = not_catch

    def info(self, message: str):
        for logger in self.loggers:
            logger.info(message)

    def debug(self, message: str):
        for logger in self.loggers:
            logger.debug(message)

    def error(self, message: str):
        for logger in self.loggers:
            logger.error(message)

    def warning(self, message: str):
        for logger in self.loggers:
            logger.warning(message)

    def success(self, message: str):
        for logger in self.loggers:
            logger.success(message)

    def catch(self, func):
        @wraps(func)
        async def wrapped(*args, **kwargs):
            new_func = func
            for logger in self.loggers[:-1]:
                new_func = logger.catch(new_func, reraise=True)
            new_func = self.loggers[-1].catch(new_func)

            return await new_func(*args, **kwargs)
        return wrapped



"""Simple and fast framework to create message brokers based microservices"""
import platform

import click


__version__ = '0.0.3.7'


def print_version(ctx: click.Context, param: click.Parameter, value: bool) -> None:
    if not value or ctx.resilient_parsing:
        return

    click.echo(
        "Running propan %s with %s %s on %s"
        % (
            __version__,
            platform.python_implementation(),
            platform.python_version(),
            platform.system(),
        )
    )
    ctx.exit()


@click.command()
@click.argument("app")
@click.option(
    "--version",
    is_flag=True,
    callback=print_version,
    expose_value=False,
    is_eager=True,
    help="Display the propan version and exit.",
)
@click.option(
    "--start",
    is_flag=True,
    default=False,
    help="Create a new propan project.",
)
@click.option(
    "--reload",
    is_flag=True,
    default=False,
    help="Reload app at code changing.",
    show_default=True,
)
@click.option(
    "--uvloop",
    default=True,
    help="Use uvloop as a default event loop.",
    show_default=True,
)
@click.option(
    "--config",
    default="config.yml",
    show_default=True,
    help="Select conf file of your consume.",
)
@click.option(
    "--workers",
    default=10,
    show_default=True,
    help="Select number of consumers.",
)
def run(
    app: str, reload: bool,
    config: str, workers: int,
    start: bool, uvloop: bool
):
    args = (app, config, workers, uvloop)

    if start:
        from propan.startproject import create
        create(app, __version__)

    if reload is True:
        from propan.supervisors.watchgodreloader import WatchGodReload
        WatchGodReload(target=_run, args=args).run()

    else:
        _run(*args)


def _run(app: str, config: str, workers: int, uvloop: bool):
    from importlib.util import spec_from_file_location, module_from_spec
    from pathlib import Path

    try:
        f, func = app.split(":", 2)

        mod_path = Path.cwd()
        for i in f.split('.'):
            mod_path = mod_path / i
        BASE_DIR = mod_path.parent.parent

        from propan.config.configuration import init_settings
        config = init_settings(
            BASE_DIR, config,
            uvloop=uvloop, 
            **{
                "MAX_CONSUMERS": workers
            }
        )
        spec = spec_from_file_location("mode", f'{mod_path}.py')
        mod = module_from_spec(spec)
        spec.loader.exec_module(mod)
        propan_app = getattr(mod, func)

    except (ValueError, FileNotFoundError, AttributeError) as e:
        from loguru import logger
        logger.error(e)
        logger.error('Please, input module like python_file:propan_app_name')
        exit()

    else:
        propan_app.run()


if __name__ == "__main__":
    run()



from typing import Protocol, Callable, Tuple


class LoggerUsecase(Protocol):
    not_catch: Tuple[Exception]

    def info(self, message: str) -> None:
        raise NotImplementedError()

    def error(self, message: str) -> None:
        raise NotImplementedError()

    def debug(self, message: str) -> None:
        raise NotImplementedError()

    def success(self, message: str) -> None:
        raise NotImplementedError()

    def warning(self, message: str) -> None:
        raise NotImplementedError()

    def catch(self, func: Callable) -> Callable:
        raise NotImplementedError()

    def _catch_base(self, func: Callable) -> Callable:
        raise NotImplementedError()


from typing import Protocol, NoReturn, Union

from .object import DbObject


class ARUsecase(Protocol):
    def create(self, item: DbObject) -> DbObject:
        raise NotImplementedError()

    def bulk_create(self, items: list[DbObject]) -> list[DbObject]:
        raise NotImplementedError()

    def update(self, item: DbObject) -> DbObject:
        raise NotImplementedError()

    def bulk_update(self, items: list[DbObject]) -> list[DbObject]:
        raise NotImplementedError()

    def update_or_create(self, item: DbObject) -> (DbObject, bool):
        raise NotImplementedError()

    def get(self, model, *args, **kwargs) -> DbObject:
        raise NotImplementedError()

    def get_or_create(self, model, *args, **kwargs) -> (DbObject, bool):
        raise NotImplementedError()

    def filter(self, model, *args, **kwargs) -> list[DbObject]:
        raise NotImplementedError()

    def delete(self, to_del: Union[DbObject, list[DbObject]]) -> NoReturn:
        raise NotImplementedError()



from typing import Optional


class MaxRetriesExceeded(Exception):
    def __init__(self, max_tries: int, message: Optional[str] = None):
        if message is None:
            message = f"Max retries '{max_tries}' exceeded"
        super().__init__(message)



from ..model.usecase import LoggerUsecase


class EmptyLogger(LoggerUsecase):
    def info(self, message: str):
        pass

    def debug(self, message: str):
        pass

    def error(self, message: str):
        pass

    def success(self, message: str):
        pass

    def warning(self, message: str):
        pass

    def catch(self, func, reraise: bool = False):
        async def wrapped(*args, **kwargs):
            return await func(*args, **kwargs)
        return wrapped



import asyncio

from propan.brokers import RabbitBroker
from propan.logger import loguru as logger

from propan.brokers.model.bus_usecase import BrokerUsecase
from propan.annotations import apply_types


class PropanApp:
    _instanse = None

    def __new__(cls, *args, **kwargs):
        if cls._instanse is None:
            cls._instanse = super().__new__(cls)
        return cls._instanse

    def __init__(
        self,
        broker: BrokerUsecase = None,
        apply_types=False,
        *args, **kwargs
    ):
        self.broker = broker or RabbitBroker(logger=logger)
        self._apply_types = apply_types
        self.loop = asyncio.get_event_loop()

    def run(self):
        try:
            self.loop.run_forever()
        finally:
            self.loop.run_until_complete(self.broker.close())

    def handle(self, queue_name: str):
        def decor(func):
            if self._apply_types:
                func = apply_types(func)
            self.loop.run_until_complete(
                self.broker.set_queue_handler(
                    queue_name=queue_name,
                    handler=func
                )
            )
        return decor



from pathlib import Path
from typing import Optional, List


BASE_DIR = Path.cwd()


def _write_file(path: Path, *content: Optional[List[str]]):
    path.touch()
    if content:
        path.write_text('\n'.join(content))


def _create_project_dir(dirname: str, version: str) -> Path:
    project_dir = BASE_DIR / dirname

    if project_dir.exists() is False:
        project_dir.mkdir()

    _write_file(project_dir / 'README.md')
    _write_file(project_dir / 'Dockerfile')

    _write_file(
        project_dir / 'docker-compose.yml',
        'services:',
        '    rabbit:',
        '       image: rabbitmq:3',
        '       environment:',
        '           - RABBITMQ_DEFAULT_USER=user',
        '           - RABBITMQ_DEFAULT_PASS=password',
        '           - RABBITMQ_DEFAULT_VHOST=en0',
        '       ports:',
        '           - 5672:5672'
    )

    _write_file(project_dir / 'requirements.txt', f'propan=={version}')

    _write_file(
        project_dir / '.gitignore', 
        'config.yml',
        ',n__pycache__',
        ',nenv/',
        ',nvenv/'
    )
    
    return project_dir


def _create_app_dir(project_dir: Path) -> Path:
    app_dir = project_dir / 'app'

    if app_dir.exists() is False:
        app_dir.mkdir()

    return app_dir


def _create_config_dir(app_dir: Path) -> Path:
    config_dir = app_dir / 'config'

    if config_dir.exists() is False:
        config_dir.mkdir()

    _write_file(
        config_dir / 'config.yml',
        'RABBIT:',
        '   HOST: 127.0.0.1',
        '   LOGIN: user',
        '   PASSWORD: password',
        '   VHOST: en0'
    )

    _write_file(
        config_dir / 'settings.py',
        'from propan.config import settings\n'
    )
    
    _write_file(config_dir / '__init__.py')
    return config_dir


def create(dirname: str, version):
    project_dir = _create_project_dir(dirname, version)
    app_dir = _create_app_dir(project_dir)
    config_dir = _create_config_dir(app_dir)

    _write_file(
        app_dir / 'serve.py',
        'from propan.app import PropanApp\n',
        'from dependencies import broker\n\n',
        'app = PropanApp(broker)\n\n',
        '@app.handle(queue_name="test")',
        'async def base_handler(message):',
        '   print(message)\n\n',
        'if __name__=="__main__":',
        '   app.run()\n'
    )

    _write_file(
        app_dir / 'dependencies.py',
        'from propan.config import settings',
        'from propan.brokers import RabbitBroker',
        'from propan.logger import loguru\n\n',
        'broker = RabbitBroker(',
        '   host=settings.RABBIT_HOST,',
        '   login=settings.RABBIT_LOGIN,',
        '   password=settings.RABBIT_PASSWORD,',
        '   vhost=settings.RABBIT_VHOST,',
        '   logger=loguru',
        ')\n'
    )

    _write_file(app_dir / '__init__.py')
    exit()



import multiprocessing
import os
import sys
from multiprocessing.context import SpawnProcess
from multiprocessing import Process
from pathlib import Path
from socket import socket
from typing import Callable, List, Optional, Union, Tuple


multiprocessing.allow_connection_pickling()
spawn = multiprocessing.get_context("spawn")


def get_subprocess(
    target: Callable[..., None], args: Tuple
) -> SpawnProcess:
    stdin_fileno: Optional[int]
    try:
        stdin_fileno = sys.stdin.fileno()
    except OSError:
        stdin_fileno = None

    kwargs = {
        "target": target,
        "args": args,
        "stdin_fileno": stdin_fileno,
    }
    return Process(target=subprocess_started, kwargs=kwargs)


def subprocess_started(
    target: Callable[..., None],
    args: Tuple,
    stdin_fileno: Optional[int],
) -> None:
    if stdin_fileno is not None:
        sys.stdin = os.fdopen(stdin_fileno)
    target(*args)


def is_dir(path: Path) -> bool:
    try:
        if not path.is_absolute():
            path = path.resolve()
        return path.is_dir()
    except OSError:
        return False


def _normalize_dirs(dirs: Union[List[str], str, None]) -> List[str]:
    if dirs is None:
        return []
    if isinstance(dirs, str):
        return [dirs]
    return list(set(dirs))


def resolve_reload_patterns(
    patterns_list: List[str], directories_list: List[str]
) -> Tuple[List[str], List[Path]]:

    directories: List[Path] = list(set(map(Path, directories_list.copy())))
    patterns: List[str] = patterns_list.copy()

    current_working_directory = Path.cwd()
    for pattern in patterns_list:
        if pattern == ".*":
            continue
        patterns.append(pattern)
        if is_dir(Path(pattern)):
            directories.append(Path(pattern))
        else:
            for match in current_working_directory.glob(pattern):
                if is_dir(match):
                    directories.append(match)

    directories = list(set(directories))
    directories = list(map(Path, directories))
    directories = list(map(lambda x: x.resolve(), directories))
    directories = list(
        set([reload_path for reload_path in directories if is_dir(reload_path)])
    )

    children = []
    for j in range(len(directories)):
        for k in range(j + 1, len(directories)):
            if directories[j] in directories[k].parents:
                children.append(directories[k])
            elif directories[k] in directories[j].parents:
                children.append(directories[j])

    directories = list(set(directories).difference(set(children)))

    return list(set(patterns)), directories


class Config:
    def __init__(
        self,
        reload_dirs: Optional[Union[List[str], str]] = [Path.cwd() / 'app'],
        reload_includes: Optional[Union[List[str], str]] = None,
        reload_excludes: Optional[Union[List[str], str]] = None
    ):
        reload_dirs = _normalize_dirs(reload_dirs)
        reload_includes = _normalize_dirs(reload_includes)
        reload_excludes = _normalize_dirs(reload_excludes)

        self.reload_includes, self.reload_dirs = resolve_reload_patterns(
            reload_includes, reload_dirs
        )

        self.reload_excludes, self.reload_dirs_excludes = resolve_reload_patterns(
            reload_excludes, []
        )

        reload_dirs_tmp = self.reload_dirs.copy()

        for directory in self.reload_dirs_excludes:
            for reload_directory in reload_dirs_tmp:
                if (
                    directory == reload_directory
                    or directory in reload_directory.parents
                ):
                    try:
                        self.reload_dirs.remove(reload_directory)
                    except ValueError:
                        pass

        for pattern in self.reload_excludes:
            if pattern in self.reload_includes:
                self.reload_includes.remove(pattern)
        if not self.reload_dirs:
            if reload_dirs:
                logger.warning(
                    "Provided reload directories %s did not contain valid "
                    + "directories, watching current working directory.",
                    reload_dirs,
                )
            self.reload_dirs = [Path.cwd() / 'app']

        print(
            f"Will watch for changes in these directories: {sorted(list(map(str, self.reload_dirs)))}",
        )



from .fetcher.adapter.async_adapter import AsyncFetcherImplementation
from .proxy.adapter.async_adapter import AsyncProxyAdapter
from .user_agent.getter import UserAgentFabric


Fetcher = AsyncFetcherImplementation
Proxy = AsyncProxyAdapter
UserAgent = UserAgentFabric

__all__ = (
    'Fetcher',
    'Proxy',
    'UserAgent'
)


import logging
import os
import signal
import threading
from socket import socket
from types import FrameType
from typing import Callable, List, Optional, Tuple

from propan.supervisors.utils import get_subprocess


HANDLED_SIGNALS = (
    signal.SIGINT,  # Unix signal 2. Sent by Ctrl+C.
    signal.SIGTERM,  # Unix signal 15. Sent by `kill <pid>`.
)


class BaseReload:
    def __init__(
        self,
        target: Callable[[Optional[List[socket]]], None],
        args: Tuple,
        reload_delay: Optional[float] = 0.5,
    ) -> None:
        self.target = target
        self.should_exit = threading.Event()
        self.pid = os.getpid()
        self.reloader_name: Optional[str] = None
        self.reload_delay = reload_delay
        self._stopped = False
        self._args = args

    def signal_handler(self, sig: signal.Signals, frame: FrameType) -> None:
        if self._stopped:
            exit()
        self._stopped = True
        self.should_exit.set()

    def run(self) -> None:
        self.startup()
        while not self.should_exit.wait(self.reload_delay):
            if self.should_restart():
                self.restart()

    def startup(self) -> None:
        message = f"Started reloader process [{self.pid}] using {self.reloader_name}"
        print(message)

        for sig in HANDLED_SIGNALS:
            signal.signal(sig, self.signal_handler)
    
        self.process = get_subprocess(
            target=self.target, args=self._args
        )
        self.process.start()

    def restart(self) -> None:
        self.process.terminate()
        self.process.join(timeout=1.0)
        print('Process successfully reloaded')
        self.process = get_subprocess(
            target=self.target, args=self._args
        )
        self.process.start()
        self._stopped = False

    def should_restart(self) -> bool:
        raise NotImplementedError("Reload strategies should override should_restart()")



from collections import Counter


class PushBackWatcher:
    memory = Counter()
    max_tries: int

    def __init__(self, max_tries: int = 3):
        self.max_tries = max_tries

    def add(self, message: str):
        self.memory[message] += 1

    def is_max(self, message: str):
        return self.memory[message] > self.max_tries

    def remove(self, message: str):
        self.memory[message] = 0
        self.memory += Counter()



from typing import NoReturn, Union, Iterable

from tortoise.exceptions import DoesNotExist
from tortoise.queryset import QuerySet

from propan.db.workers.model.recorder import ARUsecase
from propan.db.workers.model.object import DbObject


class TortoiseRepository(ARUsecase):
    async def create(self, item: DbObject) -> DbObject:
        await item.save()
        return item

    async def bulk_create(self, model, items: Iterable[DbObject]) -> list[DbObject]:
        for item in items:
            await item.save()
        return items

    async def update(self, item: DbObject) -> DbObject:
        await item.save()
        return item

    async def bulk_update(self, items: Iterable[DbObject]) -> list[DbObject]:
        for ob in items:
            await ob.save()
        return items

    async def update_or_create(self, model, defaults=None, **kwargs) -> (DbObject, bool):
        try:
            obj = await model.get(**kwargs)
            for key, value in defaults.items():
                setattr(obj, key, value)
            await obj.save()
            return obj, False

        except DoesNotExist:
            kwargs.update(defaults)
            obj = model(**kwargs)
            await obj.save()
            return obj, True

    async def get(self, model, *args, **kwargs) -> DbObject:
        return await model.get(*args, **kwargs)

    async def get_or_create(self, model, *args, **kwargs) -> (DbObject, bool):
        return await model.get_or_create(*args, **kwargs)

    async def delete(self, to_del: Union[DbObject, QuerySet]) -> NoReturn:
        await to_del.delete()

    def filter(self, model, *args, **kwargs) -> QuerySet:
        return model.filter(*args, **kwargs)


from sqlalchemy.future import select
from sqlalchemy import insert
from sqlalchemy.sql.expression import Delete, Update, Insert, bindparam

from propan.db.workers.model.model.recorder import ARUsecase
from propan.db.workers.implementation.alchemy.engine import AlchemyAsyncEngine


class AlchemyRepository(ARUsecase):
    _engine: AlchemyAsyncEngine

    def __init__(self, engine: AlchemyAsyncEngine):
        self._engine = engine

    class Decorators:
        @classmethod
        def session_mixin(cls, func):
            async def wrapped(self, *args, **kwargs):
                return await self._engine.get_session(func)(self, *args, **kwargs)
            return wrapped


    @Decorators.session_mixin
    async def delete(self, model, *args, session, **kwargs):
        stmt = Delete(model).where(*args, **kwargs)
        await session.execute(stmt)
        await session.commit()


    @Decorators.session_mixin
    async def get(self, model, *args, session, **kwargs):
        stmt = select(model).filter(*args, **kwargs)
        r = await session.execute(stmt)
        return r.scalars().fetchall()


    @Decorators.session_mixin
    async def create(self, item, session):
        session.add(item)
        await session.commit()

    @Decorators.session_mixin
    async def create_all(self, items, session):
        session.add_all(items)
        await session.commit()


    @Decorators.session_mixin
    async def update(self, items, session):
        if isinstance(items, list) is False:
            items = [items]

        model = type(items[0])
        ids = [i.id for i in items]
        fields_to_update = list(filter(lambda x: not x.startswith('_'), items[0].__dict__.keys()))

        stmt = Update(model).where(model.id == bindparam('id')).values({
            field: bindparam(field) for field in fields_to_update
        })

        await session.execute(stmt, [
            {field: item.__dict__[field] for field in fields_to_update}
            for item in items
        ])


    @Decorators.session_mixin
    async def get_or_create(self, model, *args, session, **kwargs):
        stmt = select(model).filter_by(**kwargs)
        result = await session.execute(stmt)

        if r := result.first():
            return r[0], False

        else:
            obj = model(**kwargs)
            session.add(obj)
            await session.commit()
            return obj, True


import random


class UserAgentFabric:
    @staticmethod
    def get_chrome_user_agent():
        return random.choice(USER_AGENT_CHROME)

    @staticmethod
    def get_mozila_user_agent():
        return random.choice(USER_AGENT_CHROME)

    @staticmethod
    def get_mobile_user_agent():
        return random.choice(USER_AGENT_MOBILE)


USER_AGENT_CHROME = [
    "Mozilla/5.0 (Windows NT 6.1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/88.0.4324.150 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/88.0.4324.150 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/88.0.4324.150 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.3) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/88.0.4324.150 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.3; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/88.0.4324.150 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.3; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/88.0.4324.150 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/88.0.4324.150 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/88.0.4324.150 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/88.0.4324.150 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/88.0.4324.146 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/88.0.4324.146 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/88.0.4324.146 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.3) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/88.0.4324.146 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.3; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/88.0.4324.146 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.3; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/88.0.4324.146 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/88.0.4324.146 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/88.0.4324.146 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/88.0.4324.146 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/88.0.4324.104 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/88.0.4324.104 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/88.0.4324.104 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.3) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/88.0.4324.104 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.3; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/88.0.4324.104 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.3; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/88.0.4324.104 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/88.0.4324.104 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/88.0.4324.104 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/88.0.4324.104 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/88.0.4324.96 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/88.0.4324.96 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/88.0.4324.96 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.3) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/88.0.4324.96 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.3; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/88.0.4324.96 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.3; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/88.0.4324.96 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/88.0.4324.96 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/88.0.4324.96 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/88.0.4324.96 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/87.0.4280.141 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/87.0.4280.141 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/87.0.4280.141 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.3) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/87.0.4280.141 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.3; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/87.0.4280.141 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.3; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/87.0.4280.141 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/87.0.4280.141 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/87.0.4280.141 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/87.0.4280.141 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/87.0.4280.88 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/87.0.4280.88 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/87.0.4280.88 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.3) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/87.0.4280.88 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.3; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/87.0.4280.88 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.3; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/87.0.4280.88 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/87.0.4280.88 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/87.0.4280.88 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/87.0.4280.88 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/87.0.4280.66 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/87.0.4280.66 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/87.0.4280.66 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.3) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/87.0.4280.66 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.3; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/87.0.4280.66 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.3; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/87.0.4280.66 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/87.0.4280.66 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/87.0.4280.66 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/87.0.4280.66 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/86.0.4240.198 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/86.0.4240.198 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/86.0.4240.198 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.3) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/86.0.4240.198 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.3; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/86.0.4240.198 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.3; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/86.0.4240.198 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/86.0.4240.198 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/86.0.4240.198 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/86.0.4240.198 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/86.0.4240.193 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/86.0.4240.193 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/86.0.4240.193 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.3) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/86.0.4240.193 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.3; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/86.0.4240.193 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.3; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/86.0.4240.193 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/86.0.4240.193 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/86.0.4240.193 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/86.0.4240.193 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/86.0.4240.183 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/86.0.4240.183 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/86.0.4240.183 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.3) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/86.0.4240.183 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.3; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/86.0.4240.183 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.3; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/86.0.4240.183 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/86.0.4240.183 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/86.0.4240.183 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/86.0.4240.183 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/86.0.4240.111 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/86.0.4240.111 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/86.0.4240.111 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.3) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/86.0.4240.111 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.3; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/86.0.4240.111 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.3; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/86.0.4240.111 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/86.0.4240.111 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/86.0.4240.111 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/86.0.4240.111 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/86.0.4240.75 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/86.0.4240.75 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/86.0.4240.75 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.3) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/86.0.4240.75 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.3; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/86.0.4240.75 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.3; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/86.0.4240.75 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/86.0.4240.75 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/86.0.4240.75 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/86.0.4240.75 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/85.0.4183.121 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/85.0.4183.121 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/85.0.4183.121 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.3) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/85.0.4183.121 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.3; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/85.0.4183.121 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.3; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/85.0.4183.121 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/85.0.4183.121 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/85.0.4183.121 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/85.0.4183.121 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/85.0.4183.102 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/85.0.4183.102 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/85.0.4183.102 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.3) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/85.0.4183.102 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.3; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/85.0.4183.102 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.3; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/85.0.4183.102 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/85.0.4183.102 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/85.0.4183.102 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/85.0.4183.102 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/85.0.4183.83 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/85.0.4183.83 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/85.0.4183.83 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.3) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/85.0.4183.83 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.3; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/85.0.4183.83 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.3; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/85.0.4183.83 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/85.0.4183.83 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/85.0.4183.83 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/85.0.4183.83 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.135 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.135 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.135 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.3) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.135 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.3; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.135 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.3; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.135 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.135 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.135 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.135 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.125 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.125 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.125 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.3) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.125 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.3; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.125 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.3; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.125 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.125 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.125 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.125 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.105 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.105 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.105 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.3) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.105 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.3; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.105 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.3; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.105 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.105 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.105 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.105 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.89 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.89 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.89 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.3) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.89 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.3; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.89 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.3; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.89 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.89 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.89 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.89 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/83.0.4103.116 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/83.0.4103.116 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/83.0.4103.116 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.3) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/83.0.4103.116 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.3; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/83.0.4103.116 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.3; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/83.0.4103.116 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/83.0.4103.116 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/83.0.4103.116 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/83.0.4103.116 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/83.0.4103.106 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/83.0.4103.106 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/83.0.4103.106 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.3) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/83.0.4103.106 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.3; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/83.0.4103.106 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.3; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/83.0.4103.106 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/83.0.4103.106 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/83.0.4103.106 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/83.0.4103.106 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/83.0.4103.97 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/83.0.4103.97 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/83.0.4103.97 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.3) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/83.0.4103.97 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.3; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/83.0.4103.97 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.3; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/83.0.4103.97 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/83.0.4103.97 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/83.0.4103.97 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/83.0.4103.97 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/83.0.4103.61 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/83.0.4103.61 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/83.0.4103.61 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.3) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/83.0.4103.61 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.3; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/83.0.4103.61 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.3; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/83.0.4103.61 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/83.0.4103.61 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/83.0.4103.61 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/83.0.4103.61 Safari/537.36",
]

USER_AGENT_FIREFOX = [
    "Mozilla/5.0 (Windows NT 6.1; rv:86.0) Gecko/20100101 Firefox/86.0",
    "Mozilla/5.0 (Windows NT 6.1; WOW64; rv:86.0) Gecko/20100101 Firefox/86.0",
    "Mozilla/5.0 (Windows NT 6.1; Win64; x64; rv:86.0) Gecko/20100101 Firefox/86.0",
    "Mozilla/5.0 (Windows NT 6.3; rv:86.0) Gecko/20100101 Firefox/86.0",
    "Mozilla/5.0 (Windows NT 6.3; WOW64; rv:86.0) Gecko/20100101 Firefox/86.0",
    "Mozilla/5.0 (Windows NT 6.3; Win64; x64; rv:86.0) Gecko/20100101 Firefox/86.0",
    "Mozilla/5.0 (Windows NT 10.0; rv:86.0) Gecko/20100101 Firefox/86.0",
    "Mozilla/5.0 (Windows NT 10.0; WOW64; rv:86.0) Gecko/20100101 Firefox/86.0",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:86.0) Gecko/20100101 Firefox/86.0",
    "Mozilla/5.0 (Windows NT 6.1; rv:85.0) Gecko/20100101 Firefox/85.0",
    "Mozilla/5.0 (Windows NT 6.1; WOW64; rv:85.0) Gecko/20100101 Firefox/85.0",
    "Mozilla/5.0 (Windows NT 6.1; Win64; x64; rv:85.0) Gecko/20100101 Firefox/85.0",
    "Mozilla/5.0 (Windows NT 6.3; rv:85.0) Gecko/20100101 Firefox/85.0",
    "Mozilla/5.0 (Windows NT 6.3; WOW64; rv:85.0) Gecko/20100101 Firefox/85.0",
    "Mozilla/5.0 (Windows NT 6.3; Win64; x64; rv:85.0) Gecko/20100101 Firefox/85.0",
    "Mozilla/5.0 (Windows NT 10.0; rv:85.0) Gecko/20100101 Firefox/85.0",
    "Mozilla/5.0 (Windows NT 10.0; WOW64; rv:85.0) Gecko/20100101 Firefox/85.0",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:85.0) Gecko/20100101 Firefox/85.0",
    "Mozilla/5.0 (Windows NT 6.1; rv:84.0) Gecko/20100101 Firefox/84.0",
    "Mozilla/5.0 (Windows NT 6.1; WOW64; rv:84.0) Gecko/20100101 Firefox/84.0",
    "Mozilla/5.0 (Windows NT 6.1; Win64; x64; rv:84.0) Gecko/20100101 Firefox/84.0",
    "Mozilla/5.0 (Windows NT 6.3; rv:84.0) Gecko/20100101 Firefox/84.0",
    "Mozilla/5.0 (Windows NT 6.3; WOW64; rv:84.0) Gecko/20100101 Firefox/84.0",
    "Mozilla/5.0 (Windows NT 6.3; Win64; x64; rv:84.0) Gecko/20100101 Firefox/84.0",
    "Mozilla/5.0 (Windows NT 10.0; rv:84.0) Gecko/20100101 Firefox/84.0",
    "Mozilla/5.0 (Windows NT 10.0; WOW64; rv:84.0) Gecko/20100101 Firefox/84.0",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:84.0) Gecko/20100101 Firefox/84.0",
    "Mozilla/5.0 (Windows NT 6.1; rv:83.0) Gecko/20100101 Firefox/83.0",
    "Mozilla/5.0 (Windows NT 6.1; WOW64; rv:83.0) Gecko/20100101 Firefox/83.0",
    "Mozilla/5.0 (Windows NT 6.1; Win64; x64; rv:83.0) Gecko/20100101 Firefox/83.0",
    "Mozilla/5.0 (Windows NT 6.3; rv:83.0) Gecko/20100101 Firefox/83.0",
    "Mozilla/5.0 (Windows NT 6.3; WOW64; rv:83.0) Gecko/20100101 Firefox/83.0",
    "Mozilla/5.0 (Windows NT 6.3; Win64; x64; rv:83.0) Gecko/20100101 Firefox/83.0",
    "Mozilla/5.0 (Windows NT 10.0; rv:83.0) Gecko/20100101 Firefox/83.0",
    "Mozilla/5.0 (Windows NT 10.0; WOW64; rv:83.0) Gecko/20100101 Firefox/83.0",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:83.0) Gecko/20100101 Firefox/83.0",
    "Mozilla/5.0 (Windows NT 6.1; rv:82.0) Gecko/20100101 Firefox/82.0",
    "Mozilla/5.0 (Windows NT 6.1; WOW64; rv:82.0) Gecko/20100101 Firefox/82.0",
    "Mozilla/5.0 (Windows NT 6.1; Win64; x64; rv:82.0) Gecko/20100101 Firefox/82.0",
    "Mozilla/5.0 (Windows NT 6.3; rv:82.0) Gecko/20100101 Firefox/82.0",
    "Mozilla/5.0 (Windows NT 6.3; WOW64; rv:82.0) Gecko/20100101 Firefox/82.0",
    "Mozilla/5.0 (Windows NT 6.3; Win64; x64; rv:82.0) Gecko/20100101 Firefox/82.0",
    "Mozilla/5.0 (Windows NT 10.0; rv:82.0) Gecko/20100101 Firefox/82.0",
    "Mozilla/5.0 (Windows NT 10.0; WOW64; rv:82.0) Gecko/20100101 Firefox/82.0",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:82.0) Gecko/20100101 Firefox/82.0",
    "Mozilla/5.0 (Windows NT 6.1; rv:81.0) Gecko/20100101 Firefox/81.0",
    "Mozilla/5.0 (Windows NT 6.1; WOW64; rv:81.0) Gecko/20100101 Firefox/81.0",
    "Mozilla/5.0 (Windows NT 6.1; Win64; x64; rv:81.0) Gecko/20100101 Firefox/81.0",
    "Mozilla/5.0 (Windows NT 6.3; rv:81.0) Gecko/20100101 Firefox/81.0",
    "Mozilla/5.0 (Windows NT 6.3; WOW64; rv:81.0) Gecko/20100101 Firefox/81.0",
    "Mozilla/5.0 (Windows NT 6.3; Win64; x64; rv:81.0) Gecko/20100101 Firefox/81.0",
    "Mozilla/5.0 (Windows NT 10.0; rv:81.0) Gecko/20100101 Firefox/81.0",
    "Mozilla/5.0 (Windows NT 10.0; WOW64; rv:81.0) Gecko/20100101 Firefox/81.0",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:81.0) Gecko/20100101 Firefox/81.0",
    "Mozilla/5.0 (Windows NT 6.1; rv:80.0) Gecko/20100101 Firefox/80.0",
    "Mozilla/5.0 (Windows NT 6.1; WOW64; rv:80.0) Gecko/20100101 Firefox/80.0",
    "Mozilla/5.0 (Windows NT 6.1; Win64; x64; rv:80.0) Gecko/20100101 Firefox/80.0",
    "Mozilla/5.0 (Windows NT 6.3; rv:80.0) Gecko/20100101 Firefox/80.0",
    "Mozilla/5.0 (Windows NT 6.3; WOW64; rv:80.0) Gecko/20100101 Firefox/80.0",
    "Mozilla/5.0 (Windows NT 6.3; Win64; x64; rv:80.0) Gecko/20100101 Firefox/80.0",
    "Mozilla/5.0 (Windows NT 10.0; rv:80.0) Gecko/20100101 Firefox/80.0",
    "Mozilla/5.0 (Windows NT 10.0; WOW64; rv:80.0) Gecko/20100101 Firefox/80.0",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:80.0) Gecko/20100101 Firefox/80.0",
]

USER_AGENT_MOBILE = [
    "Mozilla/5.0 (iPhone; CPU iPhone OS 14_6 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) CriOS/91.0.4472.80 Mobile/15E148 Safari/604.1",
    "Mozilla/5.0 (iPad; CPU OS 14_6 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) CriOS/91.0.4472.80 Mobile/15E148 Safari/604.1",
    "Mozilla/5.0 (iPod; CPU iPhone OS 14_6 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) CriOS/91.0.4472.80 Mobile/15E148 Safari/604.1",
    "Mozilla/5.0 (Linux; Android 10) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/91.0.4472.101 Mobile Safari/537.36",
    "Mozilla/5.0 (Linux; Android 10; SM-A205U) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/91.0.4472.101 Mobile Safari/537.36",
    "Mozilla/5.0 (Linux; Android 10; SM-A102U) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/91.0.4472.101 Mobile Safari/537.36",
    "Mozilla/5.0 (Linux; Android 10; SM-G960U) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/91.0.4472.101 Mobile Safari/537.36",
    "Mozilla/5.0 (Linux; Android 10; SM-N960U) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/91.0.4472.101 Mobile Safari/537.36",
    "Mozilla/5.0 (Linux; Android 10; LM-Q720) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/91.0.4472.101 Mobile Safari/537.36",
    "Mozilla/5.0 (Linux; Android 10; LM-X420) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/91.0.4472.101 Mobile Safari/537.36",
    "Mozilla/5.0 (Linux; Android 10; LM-Q710(FGN)) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/91.0.4472.101 Mobile Safari/537.36",
    "Mozilla/5.0 (iPhone; CPU iPhone OS 11_4 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) FxiOS/34.0 Mobile/15E148 Safari/605.1.15",
    "Mozilla/5.0 (iPad; CPU OS 11_4 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) FxiOS/34.0 Mobile/15E148 Safari/605.1.15",
    "Mozilla/5.0 (iPod touch; CPU iPhone OS 11_4 like Mac OS X) AppleWebKit/604.5.6 (KHTML, like Gecko) FxiOS/34.0 Mobile/15E148 Safari/605.1.15",
    "Mozilla/5.0 (Android 11; Mobile; rv:68.0) Gecko/68.0 Firefox/89.0",
    "Mozilla/5.0 (Android 11; Mobile; LG-M255; rv:89.0) Gecko/89.0 Firefox/89.0",
    "Mozilla/5.0 (iPhone; CPU iPhone OS 14_6 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/14.0 Mobile/15E148 Safari/604.1",
    "Mozilla/5.0 (iPad; CPU OS 14_6 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/14.0 Mobile/15E148 Safari/604.1",
    "Mozilla/5.0 (iPod touch; CPU iPhone 14_6 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/14.0 Mobile/15E148 Safari/604.1",
    "Mozilla/5.0 (Linux; Android 10; HD1913) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/91.0.4472.101 Mobile Safari/537.36 EdgA/46.3.4.5155",
    "Mozilla/5.0 (Linux; Android 10; SM-G973F) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/91.0.4472.101 Mobile Safari/537.36 EdgA/46.3.4.5155",
    "Mozilla/5.0 (Linux; Android 10; Pixel 3 XL) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/91.0.4472.101 Mobile Safari/537.36 EdgA/46.3.4.5155",
    "Mozilla/5.0 (Linux; Android 10; ONEPLUS A6003) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/91.0.4472.101 Mobile Safari/537.36 EdgA/46.3.4.5155",
    "Mozilla/5.0 (iPhone; CPU iPhone OS 14_6 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/14.0 EdgiOS/46.3.13 Mobile/15E148 Safari/605.1.15",
    "Mozilla/5.0 (Windows Mobile 10; Android 10.0; Microsoft; Lumia 950XL) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/91.0.4472.106 Mobile Safari/537.36 Edge/40.15254.603",
    "Mozilla/5.0 (Linux; Android 10; VOG-L29) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/91.0.4472.101 Mobile Safari/537.36 OPR/63.3.3216.58675",
    "Mozilla/5.0 (Linux; Android 10; SM-G970F) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/91.0.4472.101 Mobile Safari/537.36 OPR/63.3.3216.58675",
    "Mozilla/5.0 (Linux; Android 10; SM-N975F) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/91.0.4472.101 Mobile Safari/537.36 OPR/63.3.3216.58675",
    "Mozilla/5.0 (iPhone; CPU iPhone OS 14_6 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/14.0 YaBrowser/21.5.3.836 Mobile/15E148 Safari/604.1",
    "Mozilla/5.0 (iPad; CPU OS 14_6 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/14.0 YaBrowser/21.5.3.836 Mobile/15E148 Safari/605.1",
    "Mozilla/5.0 (iPod touch; CPU iPhone 14_6 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/14.0 YaBrowser/21.5.3.836 Mobile/15E148 Safari/605.1",
    "Mozilla/5.0 (Linux; arm_64; Android 11; SM-G965F) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/91.0.4472.101 YaBrowser/21.3.4.59 Mobile Safari/537.36",
]



from typing import Protocol, Callable

from propan.logger.model.usecase import LoggerUsecase

from .bus_connection import ConnectionData


class BrokerUsecase(Protocol):
    def connect(self, connection_data: ConnectionData):
        raise NotImplementedError()

    def init_channel(self) -> None:
        raise NotImplementedError()

    def publish_message(self, queue_name: str, message: str) -> None:
        raise NotImplementedError()

    def process_message(logger: LoggerUsecase, command: str) -> None:
        raise NotImplementedError()

    def set_queue_handler(
        self, queue_name: str,
        handler: Callable, retrying_on_error: bool = False
    ) -> None:
        raise NotImplementedError()

    def close(self) -> None:
        raise NotImplementedError()



from importlib.util import spec_from_file_location, module_from_spec
import os
from pathlib import Path

import sys
from typing import Generator, Tuple

import yaml
from loguru import logger

from propan.config.lazy import settings, LazySettings


def _get_recursive_name(config, name=None) -> Generator[Tuple, None, Tuple]:
    for k, v in config.items():
        if isinstance(v, dict):
            yield from _get_recursive_name(v, k)
        else:
            if name:
                k = f'{name}_{k}'
            yield k.upper(), v


def _parse_yml_config(conf_dir: Path, conffile: str = 'config.yml') -> dict:
    config = {}
    conf = conf_dir / conffile
    if conf.exists() is False:
        logger.error(f'FileNotFoundError: {conf}')
        return config

    with conf.open('r') as f:
        if (data := yaml.safe_load(f)):
            for key, value in _get_recursive_name(data):
                config[key] = os.getenv(key, default=value)

    return config


def init_settings(
    base_dir: Path,
    conffile: str = 'config.yml',
    settings_dir: str = 'app.config',
    default_settings: str = 'settings',
    uvloop: bool = True,
    **options
) -> LazySettings:
    conf_dir = base_dir
    for i in settings_dir.split('.'):
        conf_dir = conf_dir / i
    base_dir = conf_dir.parent

    config = {
        "BASE_DIR": base_dir,
        **_parse_yml_config(conf_dir, conffile),
        **options
    }
    settings.configure(None, **config)
    sys.path.append(str(base_dir))

    spec = spec_from_file_location("app", f'{conf_dir / default_settings}.py')
    mod = module_from_spec(spec)
    spec.loader.exec_module(mod)
    settings.configure(mod, **config)

    try:
        user_uv = getattr(mod, 'UVLOOP')
    except Exception:
        user_uv = True

    if uvloop and user_uv is not False:
        import uvloop
        uvloop.install()

    return settings



import os
from pathlib import Path


def get_models_ways(models_way: str) -> tuple[str]:
    dirname = _construct_dir_name(models_way)

    current_filename = __name__.split('.')[-1] + '.py'
    files = os.listdir(path=dirname)

    py_files = list(filter(lambda x: x.endswith('.py'), files))
    if "__init__.py" in py_files:
        py_files.remove('__init__.py')

    models = tuple(f'{models_way}.{file.rstrip(".py")}' for file in py_files)
    return models


def _construct_dir_name(models_way: str) -> Path:
    dirname = Path.cwd()
    for dir in models_way.split('.'):
        dirname = dirname / dir
    return dirname



from functools import wraps
import sys

from loguru import logger

from ..model.usecase import LoggerUsecase
from ..utils import find_trace


logger.remove()
logger.add(sys.stderr, format="<green>{time:DD.MM.YYYY HH:mm:ss.SSS}</green> | <cyan>{name}</cyan> | <level>{message}</level>")

def patching(record):
    record['name'] = find_trace()

logger = logger.patch(patching)


class LoguruAdapter(LoggerUsecase):
    def info(self, message: str):
        logger.info(message)

    def debug(self, message: str):
        logger.debug(message)

    def error(self, message: str):
        logger.error(message)

    def success(self, message: str):
        logger.success(message)

    def warning(self, message: str):
        logger.warning(message)

    def catch(self, func, reraise: bool = False):
        @wraps(func)
        async def wrapped(*args, **kwargs):
            return await logger.catch(
                func,
                reraise=reraise
            )(*args, **kwargs)
        return wrapped



from typing import Protocol

from propan.logger.model.usecase import LoggerUsecase
from propan.fetch.proxy.model.usecase import ProxyUsecase


class FetcherUsecase(Protocol):
    def __init__(self,
        proxy: ProxyUsecase, logger: LoggerUsecase,
        max_retries: int, sleep: int,
        timeout: int, headers: dict
    ):
        raise NotImplementedError()

    def __aenter__(self, *args, **kwargs):
        raise NotImplementedError()

    def __aexit__(self):
        raise NotImplementedError()

    def __enter__(self, *args, **kwargs):
        raise NotImplementedError()

    def __exit__(self):
        raise NotImplementedError()

    def get(self, url, *args, session, **kwargs):
        raise NotImplementedError()

    def post(self, url, *args, session, **kwargs):
        raise NotImplementedError()

    def get_json(self, url, *args, session, **kwargs):
        raise NotImplementedError()



from pathlib import Path
import traceback

from loguru import logger

from .lazy import settings
from .configuration import init_settings


SETTINGS_DIR = 'config'

def _check_conf_file(base_dir: Path) -> bool:
    conf_file = base_dir / SETTINGS_DIR / 'config.yml'
    return conf_file.exists()


base_dir = Path.cwd()
if _check_conf_file(base_dir) is False:
    called_from = traceback.StackSummary.extract(traceback.walk_stack(None))[-1]
    base_dir = Path(called_from.filename).parent

    if _check_conf_file(base_dir) is False:
        if 'python' not in str(base_dir):
            logger.error(f"There is not config file in {base_dir / SETTINGS_DIR}")
            exit()

if _check_conf_file(base_dir) is True:
    settings = init_settings(base_dir, settings_dir=SETTINGS_DIR)

__all__ = (
    'settings'
)



from typing import Protocol, Dict

from .proxie_data import FromSettings


class ProxyUsecase(Protocol):
    def __init__(self, data: FromSettings):
        raise NotImplementedError()

    def get_proxy(self) -> Dict[str, str]:
        raise NotImplementedError()

    def switch(self) -> bool:
        raise NotImplementedError()



from random import choice
import re
from typing import Union, List

import aiohttp

from propan.config import settings
from propan.fetch.proxy.model.usecase import ProxyUsecase
from propan.fetch.proxy.model.proxie_data import FromSettings


class AsyncProxyAdapter(ProxyUsecase):
    def __init__(
        self, proxy: Union[List[str], str] = settings.PROXY,
        from_settings: FromSettings = None, is_switchable=False
    ):
        try:
            self.proxy = proxy or from_settings.proxy
        except AttributeError:
            raise ValueError('Needs to set `proxy` or `from settings`')
        self.is_switchable = is_switchable

    def get_proxy(self):
        if isinstance(self.proxy, list):
            return {'https': choice(self.proxy)}
        elif isinstance(self.proxy, str):
            return {'https': self.proxy}

    async def switch(self, proxy={}) -> bool:
        if self.is_switchable:
            return await self.get_proxy()
        else:
            return False



import asyncio

from sqlalchemy.orm import sessionmaker
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.ext.asyncio import create_async_engine

from propan.db.workers.model.connection_params import Connection


class AlchemyAsyncEngine:
    async_session_maker: sessionmaker

    def __init__(self, conn: Connection):
        engine = create_async_engine(
            f"postgresql+asyncpg://{conn.user}:{conn.password}@{conn.host}/{conn.name}",
            echo=False,
        )

        self.async_session_maker = sessionmaker(
            engine, expire_on_commit=False, class_=AsyncSession
        )


    def get_session(self, func):
        async def wrapped(*args, **kwargs):
            async with self.async_session_maker() as session:
                return await func(*args, session=session, **kwargs)
        return wrapped


import asyncio
from functools import wraps
from typing import Callable, Dict

import aio_pika

from propan.config.lazy import settings

from propan.logger.model.usecase import LoggerUsecase
from propan.logger.adapter.empty import EmptyLogger

from propan.brokers.model.bus_connection import ConnectionData
from propan.brokers.model.bus_usecase import BrokerUsecase
from propan.brokers.push_back_watcher import PushBackWatcher


class AsyncRabbitQueueAdapter(BrokerUsecase):
    logger: LoggerUsecase
    _connection: aio_pika.RobustConnection
    _channel: aio_pika.RobustChannel
    _process_message: Dict[str, Callable] = {}

    def __init__(
        self,
        host=None,
        login=None,
        password=None,
        vhost=None,
        logger: LoggerUsecase = EmptyLogger(),
        connection_data=None,
        consumers=None
    ):
        self.logger = logger
        self.connection_data = connection_data or ConnectionData(
            host or settings.RABBIT_HOST, login or settings.RABBIT_LOGIN,
            password or settings.RABBIT_PASSWORD, vhost or settings.RABBIT_VHOST
        )
        self.connect()
        asyncio.get_event_loop().run_until_complete(
            self.init_channel(max_consumers=consumers or settings.MAX_CONSUMERS)
        )

    def connect(self) -> None:
        loop = asyncio.get_event_loop()
        self._connection = loop.run_until_complete(aio_pika.connect_robust(
            host=self.connection_data.host,
            login=self.connection_data.login,
            password=self.connection_data.password,
            virtualhost=self.connection_data.virtualhost,
            loop=loop
        ))

    async def init_channel(self, max_consumers: int = None) -> None:
        self._channel = await self._connection.channel()
        if max_consumers:
            await self._channel.set_qos(prefetch_count=max_consumers)

    async def set_queue_handler(
        self, queue_name: str,
        handler: Callable, retrying_on_error: bool = False
    ) -> None:
        queue = await self._channel.declare_queue(queue_name)
        self._process_message[queue_name] = handler
        self.logger.success(f'[*] Waiting for messages in {queue_name}. To exit press CTRL+C')
        await queue.consume(self.handle_message)

    async def publish_message(self, queue_name: str, message: str) -> None:
        await self._channel.default_exchange.publish(
            aio_pika.Message(str(message).encode()),
            routing_key=queue_name,
        )

    async def handle_message(self, message: aio_pika.IncomingMessage) -> None:
        queue_name = message.routing_key
        body = message.body.decode()
        async with message.process():
            self.logger.info(f"[x] Received {body} in {queue_name}")
            await self._process_message[queue_name](body)

    async def close(self):
        await self._connection.close()

    def retry(self, queue_name, try_number=3):
        watcher = PushBackWatcher(try_number)

        def decorator(func):
            @wraps(func)
            async def wrapper(message):
                try:
                    response = await func(message)

                except Exception as e:
                    watcher.add(message)
                    if not watcher.is_max(message):
                        self.logger.error(
                            f'In "{message}" error is occured. Pushing back it to rabbit.')
                        await self.publish_message(queue_name, message)
                    else:
                        self.logger.error(
                            f'"{message}" already retried {watcher.max_tries} times. Skipped.')
                        watcher.remove(message)
                    raise e

                else:
                    watcher.remove(message)
                    return response
            return wrapper
        return decorator



from loguru import logger


empty = object()

IS_CONFIGURED = False

class UserSettingsHolder:
    def __init__(self, default_settings):
        self.__dict__['_deleted'] = set()
        self.default_settings = default_settings

    def __getattr__(self, name):
        if not name.isupper() or name in self._deleted:
            raise AttributeError
        try:
            return getattr(self.default_settings, name)
        except AttributeError as e:
            logger.warning(e)
            return None

    def __setattr__(self, name, value):
        self._deleted.discard(name)
        super().__setattr__(name, value)

    def __delattr__(self, name):
        self._deleted.add(name)
        if hasattr(self, name):
            super().__delattr__(name)


class LazySettings:
    _wrapped = None

    def __init__(self):
        self._wrapped = empty

    def __getattr__(self, name):
        if self._wrapped is empty:
            self._setup(name)
        val = getattr(self._wrapped, name)
        self.__dict__[name] = val
        return val

    def __setattr__(self, name, value):
        if name == '_wrapped':
            self.__dict__.clear()
        else:
            self.__dict__.pop(name, None)
        super().__setattr__(name, value)

    def __delattr__(self, name):
        super().__delattr__(name)
        self.__dict__.pop(name, None)

    def configure(self, default_settings, **options):
        global IS_CONFIGURED
        if IS_CONFIGURED is True:
            return self._wrapped

        holder = UserSettingsHolder(default_settings)
        for name, value in options.items():
            if not name.isupper():
                raise TypeError('Setting %r must be uppercase.' % name)
            setattr(holder, name, value)

        self._wrapped = holder
        if default_settings is not None:
            IS_CONFIGURED = True

settings = LazySettings()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['propan',
 'propan.annotations',
 'propan.brokers',
 'propan.brokers.adapter',
 'propan.brokers.model',
 'propan.config',
 'propan.db',
 'propan.db.workers',
 'propan.db.workers.implementation',
 'propan.db.workers.implementation.alchemy',
 'propan.db.workers.implementation.tortoise',
 'propan.db.workers.model',
 'propan.fetch',
 'propan.fetch.fetcher',
 'propan.fetch.fetcher.adapter',
 'propan.fetch.fetcher.model',
 'propan.fetch.proxy',
 'propan.fetch.proxy.adapter',
 'propan.fetch.proxy.model',
 'propan.fetch.user_agent',
 'propan.logger',
 'propan.logger.adapter',
 'propan.logger.model',
 'propan.supervisors']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=5.4.1,<6.0.0',
 'aiohttp>=3.7.4.post0,<4.0.0',
 'click',
 'loguru',
 'pydantic>=1.8.2,<2.0.0',
 'watchgod']

extras_require = \
{'rabbit': ['aio-pika>=6.8.0,<7.0.0']}

entry_points = \
{'console_scripts': ['propan = propan:run']}

setup_kwargs = {
    'name': 'propan',
    'version': '0.0.3.7',
    'description': 'Simple and fast framework to create message brokers based microservices',
    'long_description': 'Propan существует, чтобы максимально упростить для вас создание микросервисов вокруг брокеров сообщений (например, RabbitMQ)\n\nОсновные особенности:\n\n* **Окружение**: максимальное удоство для работы с настройками проекта.\n* **Скорость разработки**: один декоратор - один потребитель на очередь.\n* **Обработка входных параметров**: автоматическая сериализация сообщений из RabbitMQ в соответствии с аннотацией типов.\n* **Расширяемость**: возможность создавать собственные классы-реализации интерфейсов Propan.\n* **Инъекция зависимостей**: единый подход к работе с зависимостям во всем проекте.\n\n## Окружение\n\nPython 3.8+\n\n\n## Примеры\n\n### Создание проекта\n\n* Создайте проект с помощью `propan --start test_project`\n* Перейдите в директорию проекта `cd test_project`\n* Создайте `app/config/settings.py` с данными для подключения к RabbitMQ\n\n```Python\nRABBIT_HOST = "127.0.0.1"\nRABBIT_LOGIN = "guest"\nRABBIT_PASSWORD = "guest"\nRABBIT_VIRTUALHOST = "/"\n```\n\n* Настройки проекта также можно указывать в `app/config/config.yml` файле\n\n```yml\nRABBIT:\n    host: \'127.0.0.1\'\n    login: \'guest\'\n    password: \'guest\'\n    vhost: /\n```\n\nВ таком случае переменные из `.yml` файла будут транслированы в настройки в верхнем регистре\n* Данные поля для RabbitMQ автоматически используются как параметры по умолчанию\n\n```yml\nRabbit:\n    host: \'127.0.0.1\'\n```\nАналогичен\n```yml\nRABBIT_HOST: \'127.0.0.1\'\n```\n\nПроект также автоматически берет переменные из окружения при наличии совпадений в названии переменных из `.yml` файла\n* Таким образом приоритеты использования переменных в конфликтных ситуациях расставлены следующим образом: `environment` > `config.yml` > `settings.py`\n* При попытке обратиться к переменной, которой нет в настройках, propan выдаст warning и `None` в качестве значения\n* `.yml` файлы можно переключать при запуске с помощью флага `--config=prod.yml` или `-C prod.yml`\n    * в таком случае все используемые `.yml` файлы должны находиться в директории `app/config/`\n* если вы не хотите использовать `uvloop` в качестве event loop\'а по умолчанию, укажите это в `settings.py`\n```Python\nUVLOOP = False\n```\n\n* После указания данных для подключения к RabbitMQ создадим коннектор в `app/dependencies.py`\n\n```Python\nimport asyncio\n\nfrom propan.config import settings\nfrom propan.brokers import RabbitAdapter\n\nqueue_adapter = RabbitAdapter(\n    host = settings.RABBIT_HOST,\n    login = settings.RABBIT_LOGIN,\n    password = settings.RABBIT_PASSWORD,\n    virtualhost = settings.RABBIT_VHOST,\n    max_consumers=settings.MAX_CONSUMERS,\n) # данные поля settings используются для инициализации по умолчанию\n```\nГлобальные найстроки используются во всех случаях, когда необходимо получить доступ к константам проекта (инициализируются, как указано выше, при старте приложения)\n\n```Python\nfrom propan.config import settings\n```\n\n`settings.MAX_CONSUMERS` указывается при запуске проекта с помощью флага `--workers=10` или `-W 10` (10-значение по умолчанию) и определяет допустимое количество одновременно обрабатываемых сообщений\n\n* После создания `queue_adapter` создадим `PropanApp` в `app/serve.py`\n\n```Python\nfrom propan.app import PropanApp\n\nfrom .dependencies import queue_adapter\n\n# broker по умолчанию - RabbitAdapter(logger=loguru)\napp = PropanApp(broker=queue_adapter)\n\n@app.handle(queue_name="test_queue")\nasync def base_handler(message):\n    print(message)\n```\n\n* Запуск проекта осуществляется с помощью команды `propan app.serve:app`\n    * `app.serve` - путь к файлу serve, а `app` - название экземпляра `PropanApp` в коде.\n    * Запуск по умолчанию: `propan app.serve:app -C config.yml -W 10`\n    * Используйте флаг `--reload` для запуска проекта в тестовом режиме с автоматической перезагрузкой при изменении файлов\n\n\n\n### Сериализация входных параметров\n\n* Так как входящие значения RabbitMQ представляют из себя строку, для сериализации входных параметров Propan использует аннотацию типов\n\n```Python\nfrom propan.app import PropanApp\n\nfrom .dependencies import queue_adapter\n\napp = PropanApp(\n    broker=queue_adapter,\n    apply_types=True\n)\n\n@app.handle(queue_name="test_queue")\nasync def base_handler(user_id: int):  # приведение входного значения к типу int\n    print(message)\n```\n\nИспользование глобальной опции `apply_types=True` включит сериализацию входных параметров для всех handler\'ов приложения\n\n* Для использования сериализации в отдельных handler\'ах необходимо использовать специальный декоратор `apply_types`\n\n```Python\nfrom propan.app import PropanApp\nfrom propan.annotations import apply_types\n\nfrom .dependencies import queue_adapter\n\napp = PropanApp(broker=queue_adapter)\n\n@app.handle(queue_name="test_queue")\n@apply_types\nasync def base_handler(user_id: int):\n    print(message)\n```\n\n* Замечания\n    * функция-handler всегда должна принимать на вход один аргумент\n\n* Для сериализации более сложных объектов возможно использование классов-оберток над `pydantic`\n\n```Python\nfrom typing import Optional\n\nfrom propan.app import PropanApp\nfrom propan.annotations import MessageModel\n\napp = PropanApp(\n    broker=queue_adapter,\n    apply_types=True\n)\n\nclass User(MessageModel):\n    username: str\n    user_id: Optional[int]\n\n@app.handle(queue_name="test_queue")\nasync def base_handler(user: User):\n    print(user)\n```\n\n\n### Логирование\n\nВсе классы Propan, выводящие служебную информацию требуют для этого любой экземпляр logger\'a, являющегося реализацией интерфейса `propan.logger.model.usecase.LoggerUsecase`.\nРекомендуется использовать экземпляр этого класса во всем проекте, инициализируя его в `app/dependencies.py`.\n\n```Python\nfrom propan.logger.adapter.loguru_usecase import LoguruAdapter\n# также можно использовать\nfrom propan.logger import loguru # является экземпляром LoguruAdapter\n\nlogger = LoguruAdapter()\nqueue_adapter = RabbitAdapter(\n    logger=logger\n)\n```\n\n* По умолчанию во всех классах Propan используется `propan.logger.adapter.empty.EmptyLogger`\n\nТакже возможно построение цепочки обработки logger\'ов путем использования в качестве экземпляра logger класса `propan.logger.LoggerSimpleComposition`\n\n```Python\nfrom propan.logger import LoggerSimpleComposition, loguru\n\nlogger = LoggerSimpleComposition(\n    loguru, loguru\n) # последовательнок применение двух логгеров loguru\n```\n\n* В таком случае logger\'ы будут применяться в том порядке, в котором они были переданы в конструктор\n\nВсе logger\'ы поддерживают декоратор `logger.catch`, который позволяет совершать какие-либо действия при возникновении ошибки в функции\n\n```Python\n@logger.catch\n@app.handle(queue_name="test_queue")\nasync def base_handler(user: str):\n    print(user)\n```\n\n\n### Дополнительно\n\nДля возвращения сообщения в очередь при возникновении ошибки c отслеживанием количества попыток повторной обработки используйте декоратор `broker.retry`\n\n```Python\n@app.broker.retry(queue_name="test_queue", try_number=3)\n@app.handle(queue_name="test_queue")\nasync def base_handler(user: str):\n    print(user)\n```\n\n* Замечания\n    * При превышении количества повторных попыток `broker` вызывает метод `error` своего экземпляра `logger`\'а, а сообщение извлекается из очереди\n    * По умолчанию количество попыток - 3, название очереди - обязательный аргумент\n\nВ таком случае также может быть полезным использование декоратора `ignore_exceprions`\n\n```Python\nfrom propan.logger import ignore_exceptions\n\nNOT_CATCH = (ValueError,)\n\n@app.broker.retry(queue_name="test_queue", try_number=3)\n@app.handle(queue_name="test_queue")\n@ignore_exceptions(logger, NOT_CATCH)\nasync def base_handler(user: str):\n    print(user)\n```\n\n* Ошибки из `NOT_CATCH` не будут передаваться дальше по стеку, а будут обработаны с помощью метода `error` переданного экземпляра `logger`\'а\n\n\n\nХудший вариант вашего приложения будет выглядеть следующим образом:\n\n```Python\nfrom propan.app import PropanApp\nfrom propan.annotations import apply_types\nfrom propan.logger import ignore_exceptions\n\nfrom .dependencies import queue_adapter, logger\n\napp = PropanApp(\n    broker=queue_adapter\n)\n\nNOT_CATCH = (ValueError,)\n\n@logger.catch\n@app.broker.retry(queue_name="test_queue", try_number=3)\n@app.handle(queue_name="test_queue")\n@ignore_exceptions(logger, NOT_CATCH)\n@apply_types\nasync def base_handler(user: str):\n    print(user)\n```\n\n# Удачи!\n',
    'author': 'PasNA6713',
    'author_email': 'diementros@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/PasNA6713/Propan',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>3.8,<4',
}


setup(**setup_kwargs)



from functools import wraps
from inspect import getfullargspec


def apply_types(func):
    f_info = getfullargspec(func)
    arg_names = f_info.args
    annotations = f_info.annotations

    def _cast_type(arg_value, arg_name):
        if (arg_type := annotations.get(arg_name)) is not None and \
                isinstance(arg_value, arg_type) is False:
            arg_value = arg_type(arg_value)
        return arg_value

    @wraps(func)
    def wrapper(*args, **kwargs):
        kw = {
            arg_name: _cast_type(arg_value, arg_name)
            for arg_name, arg_value in
            (*zip(arg_names, args), *kwargs.items())
        }
        return _cast_type(func(**kw), "return")
    return wrapper



from dataclasses import dataclass
from typing import Union, List


@dataclass
class FromSettings:
    proxy: Union[str, List[str]]

    def __init__(self, data):
        if not isinstance(data, str):
            if not isinstance(data, list):
                raise ValueError('Proxy must be str or list[str] type')
            elif not isinstance(data[0], str):
                raise ValueError('Proxy must be str or list[str] type')
        self.proxy = data



from itertools import dropwhile
from functools import wraps
import traceback
from typing import Iterable, Callable

from propan.logger.model.usecase import LoggerUsecase
from propan.config.lazy import settings


def ignore_exceptions(logger: LoggerUsecase, ignored: Iterable[Exception]):
    def decorator(func: Callable):
        @wraps(func)
        async def wrapper(*args, **kwargs):
            try:
                return await func(*args, **kwargs)
            except ignored as e:
                logger.error(e)
        return wrapper
    return decorator


def find_trace() -> str:
    file = catch_call()
    filepath = file.filename.split(f'{settings.BASE_DIR}/')
    if len(filepath) == 1:
        path = filepath[0]
    else:
        path = filepath[1]
    path = path.rstrip('.py').replace('/', '.')
    func_name = file.name
    line = file.lineno
    if 'propan' in path:
        file = path.split(".")[-1]
        return f'{file}:{func_name}:{line}'
    return f'{path}:{func_name}:{line}'


def catch_call(ignore='logger') -> traceback.FrameSummary:
    # получаем пути файлов из стека вызовов
    trace = traceback.StackSummary.extract(traceback.walk_stack(None))
    # отсеиваем все вызовы из директории "logger" и берем первый снаружи
    called_from = list(dropwhile(lambda x: ignore in x.filename, trace))[0]
    return called_from



import asyncio
from json.decoder import JSONDecodeError
from typing import Optional

import aiohttp
from aiohttp.client_exceptions import (
    ServerTimeoutError, ClientHttpProxyError,
    ClientProxyConnectionError, ContentTypeError,
    ClientOSError, ServerDisconnectedError
)

from propan.logger.model.usecase import LoggerUsecase
from propan.logger.adapter.empty import EmptyLogger

from propan.fetch.proxy.model.usecase import ProxyUsecase
from propan.fetch.fetcher.model.usecase import FetcherUsecase
from propan.fetch.fetcher.model.errors import MaxRetriesExceeded


def session_mixin(func):
    async def wrapper(self, url, *args, **kwargs):
        for try_number in range(self._max_retries):

            try:
                if session := kwargs.pop('session', self._session):
                    return await func(self, url, *args, session=session, proxy=self._proxy, **kwargs)

                else:
                    async with aiohttp.ClientSession() as session:
                        self._proxy = self.proxy_adapter.get_proxy()['https'] if self.proxy_adapter else None
                        return await func(self, url, *args, session=session, proxy=self._proxy, **kwargs)

            except (ServerTimeoutError, asyncio.exceptions.TimeoutError, ServerDisconnectedError):
                self.logger.error(f'Timeout error {url}')
                await self.sleep()
                continue

            except (ClientHttpProxyError, ClientProxyConnectionError):
                self.logger.error(f'Proxy error {url}')
                await self.sleep()
                continue

            except (ContentTypeError, JSONDecodeError):
                self.logger.error(f'Сработала защита от ботов {url}')
                is_proxy_switched = await self.proxy_adapter.switch(self._proxy)
                if is_proxy_switched is False:
                    await self.sleep()
                continue

        raise MaxRetriesExceeded(self._max_retries)
    return wrapper


class AsyncFetcherImplementation(FetcherUsecase):
    proxy_adapter: Optional[ProxyUsecase]
    logger: LoggerUsecase

    _max_retries: int
    _sleep_time: int
    _timeout_time: int
    _headers: dict

    _proxy: Optional[str] = None
    _session: Optional[aiohttp.ClientSession] = None

    def __init__(self,
        logger: LoggerUsecase = EmptyLogger(), proxy_adapter: ProxyUsecase = None,
        max_retries=3, sleep=5,
        timeout=10, headers={},
    ):
        self.logger = logger
        self.proxy_adapter = proxy_adapter
        self._max_retries = max_retries
        self._sleep_time = sleep
        self._timeout_time = timeout
        self._headers = headers

    async def __aenter__(self, *args, **kwargs):
        self._proxy = self.proxy_adapter.get_proxy()['https'] if self.proxy_adapter else None
        self._session = aiohttp.ClientSession(*args, **kwargs)
        return self._session

    async def __aexit__(self, *args, **kwargs):
        if self._session is not None:
            await self._session.close()
        self._session = None
        self._proxy = None

    @session_mixin
    async def get(self, url, *args, session, **kwargs):
        async with session.get(
            url, *args,
            headers=kwargs.pop('headers', self._headers),
            timeout=kwargs.pop('timeout', self._timeout_time),
            **kwargs
        ) as response:
            return response

    @session_mixin
    async def post(self, url, *args, session, **kwargs):
        async with session.post(
            url, *args,
            headers=kwargs.pop('headers', self._headers),
            timeout=kwargs.pop('timeout', self._timeout_time),
            **kwargs
        ) as response:
            return response

    @session_mixin
    async def get_json(self, url, *args, session, check_content_type: bool = True, **kwargs):
        async with session.get(
            url, *args,
            headers=kwargs.pop('headers', self._headers),
            timeout=kwargs.pop('timeout', self._timeout_time),
            **kwargs
        ) as response:
            if check_content_type is True:
                try:
                    return await response.json()
                except (ContentTypeError, JSONDecodeError) as e:
                    if 'text/plain; charset=utf-8' in e.message:
                        self.logger.error('Wrong server response header. Needs to use "check_content_type=False" option')
                        return
                    else:
                        raise e
            else:
                return await response.json(content_type=None)

    async def sleep(self):
        await asyncio.sleep(self._sleep_time)


from .adapter.loguru_usecase import LoguruAdapter
from .composition import LoggerSimpleComposition
from .utils import ignore_exceptions


loguru = LoguruAdapter()

__all__ = (
    'loguru',
    'ignore_exceptions',
    'LoggerSimpleComposition'
)
